import pytest


@pytest.mark.parametrize('number_of_nodes', [2])
def test_stake_and_retrieve(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    initial_staked_amount = fair.staking.get_staked_amount()
    initial_staked_nodes = fair.staking.get_staked_nodes()
    initial_node_share = fair.staking.get_node_share(node_id)

    assert isinstance(initial_staked_amount, int)
    assert isinstance(initial_staked_nodes, list)
    assert isinstance(initial_node_share, int)
    assert initial_staked_amount >= 0
    if initial_staked_amount == 0:
        assert len(initial_staked_nodes) == 0
    else:
        assert len(initial_staked_nodes) > 0


@pytest.mark.parametrize('number_of_nodes', [2])
def test_stake_to_node(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    staked_amount = fair.staking.get_staked_amount()
    staked_nodes = fair.staking.get_staked_nodes()
    node_share = fair.staking.get_node_share(node_id)

    assert isinstance(staked_amount, int)
    assert isinstance(staked_nodes, list)
    assert isinstance(node_share, int)


@pytest.mark.parametrize('number_of_nodes', [2])
def test_get_staked_amount_for_holder(fair, fair_active_nodes):
    holder_address = fair_active_nodes[0].address

    staked_amount = fair.staking.get_staked_amount_for(holder_address)
    staked_nodes = fair.staking.get_staked_nodes_for(holder_address)

    assert isinstance(staked_amount, int)
    assert isinstance(staked_nodes, list)
    assert staked_amount >= 0
    assert all(isinstance(node_id, int) for node_id in staked_nodes)


@pytest.mark.parametrize('number_of_nodes', [2])
def test_get_staked_to_node_amount_for(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id
    holder_address = fair_active_nodes[1].address

    staked_to_node = fair.staking.get_staked_to_node_amount_for(node_id, holder_address)

    assert isinstance(staked_to_node, int)
    assert staked_to_node >= 0


@pytest.mark.parametrize('number_of_nodes', [1])
def test_fee_related_functions(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    earned_fee = fair.staking.get_earned_fee_amount(node_id)

    assert isinstance(earned_fee, int)
    assert earned_fee >= 0


@pytest.mark.parametrize('number_of_nodes', [1])
def test_node_share_calculation(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    node_share = fair.staking.get_node_share(node_id)

    assert isinstance(node_share, int)
    assert node_share >= 0


def test_staking_view_functions_with_no_nodes(fair):
    staked_amount = fair.staking.get_staked_amount()
    staked_nodes = fair.staking.get_staked_nodes()

    assert isinstance(staked_amount, int)
    assert isinstance(staked_nodes, list)
    assert staked_amount >= 0
    if staked_amount == 0:
        assert len(staked_nodes) == 0
    else:
        assert len(staked_nodes) > 0


@pytest.mark.parametrize('number_of_nodes', [2])
def test_get_staked_to_node_amount(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    staked_to_node = fair.staking.get_staked_to_node_amount(node_id)

    assert isinstance(staked_to_node, int)
    assert staked_to_node >= 0


@pytest.mark.parametrize('number_of_nodes', [1])
def test_get_node_fee_rate(fair, fair_active_nodes):
    node = fair.nodes.get_by_address(fair_active_nodes[0].address)
    node_id = node.id

    fee_rate = fair.staking.get_node_fee_rate(node_id)

    assert isinstance(fee_rate, int)
    assert fee_rate >= 0


def test_staking_public_methods_present(fair):
    staking = fair.staking
    public_attrs = [
        'add_allowed_receiver',
        'remove_allowed_receiver',
        'request_send_all_fees',
        'request_all_fees',
        'set_fee_rate',
        'request_fees',
        'request_send_fees',
        'get_earned_fee_amount',
        'request_retrieve',
        'request_retrieve_all',
        'claim_request',
        'set_self_stake_requirement',
        'set_retrieving_delay',
        'get_total_in_exit_queue',
        'get_retrieving_delay',
        'get_exit_requests_count_for',
        'get_my_exit_requests_count',
        'get_exit_request',
        'get_exit_request_at',
        'get_exit_requests_for',
        'get_unlocked_exit_request_for',
        'is_request_unlocked',
        'stake',
    ]
    for attr in public_attrs:
        assert hasattr(staking, attr)


@pytest.mark.parametrize('number_of_nodes', [1])
def test_self_stake_requirement_update(fair, fair_active_nodes):
    staking = fair.staking
    original = staking.self_stake_requirement()
    assert isinstance(original, int)
    new_value = original + 1 if original > 0 else 2
    staking.set_self_stake_requirement(new_value)
    updated = staking.self_stake_requirement()
    assert updated == new_value


@pytest.mark.parametrize('number_of_nodes', [1])
def test_exit_queue_getters_initial(fair, fair_active_nodes):
    staking = fair.staking
    node_id = fair.nodes.get_by_address(fair_active_nodes[0].address).id
    assert isinstance(staking.get_retrieving_delay(), int)
    total_exit_queue = staking.get_total_in_exit_queue(staking.skale.wallet.address)
    assert isinstance(total_exit_queue, int)
    assert isinstance(staking.is_within_stake_limit(node_id), bool)
    assert staking.get_my_exit_requests_count() >= 0
    assert total_exit_queue >= 0



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from eth_typing import ChecksumAddress
from web3.contract.contract import ContractFunction

from skale.contracts.base_contract import transaction_method
from skale.contracts.skale_contract import SkaleContract


class Etherbase(SkaleContract):
    """Etherbase contract"""

    def default_admin_role(self) -> bytes:
        return self.contract.functions.DEFAULT_ADMIN_ROLE().call()

    def ether_manager_role(self) -> bytes:
        return self.contract.functions.ETHER_MANAGER_ROLE().call()

    @transaction_method
    def retrieve(self, address: ChecksumAddress) -> ContractFunction:
        return self.contract.functions.retrieve(address)

    @transaction_method
    def partially_retrieve(self, address: ChecksumAddress, amount: int) -> ContractFunction:
        return self.contract.functions.partiallyRetrieve(address, amount)

    def has_role(self, role: bytes, address: ChecksumAddress) -> bool:
        return bool(self.contract.functions.hasRole(role, address).call())

    def get_role_admin(self, role: bytes) -> bytes:
        return self.contract.functions.getRoleAdmin(role).call()

    def get_role_member(self, role: bytes, index: int) -> bytes:
        return self.contract.functions.getRoleMember(role, index).call()

    def get_role_member_count(self, role: bytes) -> int:
        return self.contract.functions.getRoleMemberCount(role).call()

    @transaction_method
    def grant_role(self, role: bytes, address: ChecksumAddress) -> ContractFunction:
        return self.contract.functions.grantRole(role, address)

    @transaction_method
    def revoke_role(self, role: bytes, address: ChecksumAddress) -> ContractFunction:
        return self.contract.functions.revokeRole(role, address)

    @transaction_method
    def renounce_role(self, role: bytes, address: ChecksumAddress) -> ContractFunction:
        return self.contract.functions.renounceRole(role, address)

    def get_version(self) -> str:
        return self.contract.functions.version().call()

    @transaction_method
    def set_version(self, new_version) -> ContractFunction:
        return self.contract.functions.setVersion(new_version)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging
import time
from typing import Any, Dict, Iterable
from urllib.parse import urlparse

import requests
from eth_keys.main import lazy_key_api as keys
from eth_typing import Address, AnyAddress, ChecksumAddress, HexStr
from requests.exceptions import ConnectionError  # type: ignore
from web3 import HTTPProvider, LegacyWebSocketProvider, Web3
from web3.exceptions import ProviderConnectionError, StaleBlockchain, TransactionNotFound
from web3.middleware import AttributeDictMiddleware, Middleware, StalecheckMiddlewareBuilder
from web3.providers.base import JSONBaseProvider
from web3.types import ENS, Nonce, TxReceipt, _Hash32

import skale.config as config
from skale.transactions.exceptions import TransactionFailedError, TransactionNotMinedError
from skale.utils.cache import RedisCacheConfig, redis_cache_middleware
from skale.utils.constants import GAS_PRICE_COEFFICIENT
from skale.utils.exceptions import NoSyncedEndpointError

logger = logging.getLogger(__name__)


WS_MAX_MESSAGE_DATA_BYTES = 5 * 1024 * 1024
MAX_WAITING_TIME = 3 * 60 * 60  # 3 hours
BLOCK_WAITING_TIMEOUT = 1
DEFAULT_HTTP_TIMEOUT = 120
DEFAULT_BLOCKS_TO_WAIT = 50


def get_provider(
    endpoint: str,
    timeout: int = DEFAULT_HTTP_TIMEOUT,
    request_kwargs: Dict[str, Any] | None = None,
    session: Any = None,
) -> JSONBaseProvider:
    scheme = urlparse(endpoint).scheme
    if scheme == 'ws' or scheme == 'wss':
        kwargs = request_kwargs or {'max_size': WS_MAX_MESSAGE_DATA_BYTES}
        return LegacyWebSocketProvider(endpoint, websocket_timeout=timeout, websocket_kwargs=kwargs)

    if scheme == 'http' or scheme == 'https':
        kwargs = {'timeout': timeout, **(request_kwargs or {})}
        return HTTPProvider(endpoint, session=session, request_kwargs=kwargs)

    raise Exception('Wrong endpoint option.Supported endpoint schemes: http/https/ws/wss')


def init_web3(
    endpoint: str,
    provider_timeout: int = DEFAULT_HTTP_TIMEOUT,
    middlewares: Iterable[Middleware] | None = None,
    cache_config: RedisCacheConfig | None = None,
    ts_diff: int | None = None,
    session: requests.Session | None = None,
) -> Web3:
    provider = get_provider(
        endpoint,
        timeout=provider_timeout,
        session=session,
    )
    provider.cache_allowed_requests = True
    w3 = Web3(provider)
    if not middlewares:
        ts_diff = ts_diff or config.ALLOWED_TS_DIFF
        if not ts_diff == config.NO_SYNC_TS_DIFF:
            stalecheck_middleware = StalecheckMiddlewareBuilder.build(ts_diff)
            middlewares = [stalecheck_middleware, AttributeDictMiddleware]
        else:
            middlewares = [AttributeDictMiddleware]

    if cache_config is not None:
        caching_middleware = redis_cache_middleware(cache_config)
        middlewares_list = list(middlewares)
        try:
            attribute_index = middlewares_list.index(AttributeDictMiddleware)
        except ValueError:
            attribute_index = len(middlewares_list)
        middlewares_list.insert(attribute_index, caching_middleware)
        middlewares = middlewares_list

    for middleware in middlewares:
        w3.middleware_onion.add(middleware)
    return w3


def get_endpoint(
    endpoint: str | list[str],
    provider_timeout: int = DEFAULT_HTTP_TIMEOUT,
    ts_diff: int | None = None,
) -> str:
    if isinstance(endpoint, str):
        return endpoint
    elif isinstance(endpoint, list) and len(endpoint) > 0:
        return _get_connected_endpoint(endpoint, provider_timeout=provider_timeout, ts_diff=ts_diff)
    else:
        raise ValueError('Endpoint must be a string or a non-empty list of strings.')


def _get_connected_endpoint(
    endpoints: list[str], provider_timeout: int, ts_diff: int | None
) -> str:
    for url in endpoints:
        try:
            w3 = init_web3(url, provider_timeout=provider_timeout, ts_diff=ts_diff)
            if w3.eth.block_number:
                return url
        except (ProviderConnectionError, ConnectionError) as e:
            logger.warning(f'Could not connect to {url}. Error: {e}. Trying next endpoint...')
            time.sleep(2)
        except StaleBlockchain as e:
            logger.warning(f'Endpoint {url} is out of sync. Error: {e}. Trying next endpoint...')
            time.sleep(2)
    raise NoSyncedEndpointError(f'Could not connect to any RPC endpoints: {endpoints}')


def get_receipt(web3: Web3, tx: _Hash32) -> TxReceipt:
    return web3.eth.get_transaction_receipt(tx)


def get_eth_nonce(web3: Web3, address: Address | ChecksumAddress | ENS) -> Nonce:
    return web3.eth.get_transaction_count(address)


def wait_for_receipt_by_blocks(
    web3: Web3,
    tx: _Hash32,
    blocks_to_wait: int = DEFAULT_BLOCKS_TO_WAIT,
    timeout: int = MAX_WAITING_TIME,
) -> TxReceipt:
    blocks_to_wait = blocks_to_wait or DEFAULT_BLOCKS_TO_WAIT
    timeout = timeout or MAX_WAITING_TIME
    previous_block = web3.eth.block_number
    current_block = previous_block
    wait_start_time = time.time()
    while (
        time.time() - wait_start_time < timeout and current_block <= previous_block + blocks_to_wait
    ):
        try:
            receipt = get_receipt(web3, tx)
        except TransactionNotFound:
            receipt = None
        if receipt is not None:
            return receipt
        current_block = web3.eth.block_number
        time.sleep(3)
    raise TransactionNotMinedError(
        f'Transaction with hash: {str(tx)} not found in {blocks_to_wait} blocks.'
    )


def wait_receipt(web3: Web3, tx: _Hash32, retries: int = 30, timeout: int = 5) -> TxReceipt:
    for _ in range(0, retries):
        try:
            receipt = get_receipt(web3, tx)
        except TransactionNotFound:
            receipt = None
        if receipt is not None:
            return receipt
        time.sleep(timeout)  # pragma: no cover
    raise TransactionNotMinedError(
        f'Transaction with hash: {str(tx)} not mined after {retries} retries.'
    )


def check_receipt(receipt: TxReceipt, raise_error: bool = True) -> bool:
    if receipt['status'] != 1:  # pragma: no cover
        if raise_error:
            raise TransactionFailedError(f'Transaction failed, see receipt {receipt}')
        else:
            return False
    return True


def wait_for_confirmation_blocks(
    web3: Web3, blocks_to_wait: int, timeout: int = MAX_WAITING_TIME, request_timeout: int = 5
) -> None:
    current_block = start_block = web3.eth.block_number
    logger.info(
        f'Current block number is {current_block}, '
        f'waiting for {blocks_to_wait} confimration blocks to be mined'
    )
    wait_start_time = time.time()
    while time.time() - wait_start_time < timeout and current_block <= start_block + blocks_to_wait:
        current_block = web3.eth.block_number
        time.sleep(request_timeout)


def private_key_to_public(pr: HexStr) -> HexStr:
    pr_bytes = Web3.to_bytes(hexstr=pr)
    prk = keys.PrivateKey(pr_bytes)
    pk = prk.public_key
    return HexStr(pk.to_hex())


def public_key_to_address(pk: HexStr) -> HexStr:
    hash = Web3.keccak(hexstr=str(pk))
    return Web3.to_hex(hash[-20:])


def private_key_to_address(pr: HexStr) -> HexStr:
    pk = private_key_to_public(pr)
    return public_key_to_address(pk)


def to_checksum_address(address: AnyAddress | str | bytes) -> ChecksumAddress:
    return Web3.to_checksum_address(address)


def default_gas_price(web3: Web3) -> int:
    return web3.eth.gas_price * GAS_PRICE_COEFFICIENT



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from typing import List, Tuple

from eth_typing import ChecksumAddress
from web3 import Web3
from web3.types import RPCEndpoint

from skale import SkaleManager
from skale.dataclasses.schain_options import SchainOptions
from skale.transactions.result import TxRes
from skale.types.node import NodeId, NodeStatus
from skale.types.schain import SchainName
from skale.types.validator import ValidatorId
from skale.utils.contracts_provision import (
    D_DELEGATION_INFO,
    D_DELEGATION_PERIOD,
    D_STAKE_MULTIPLIER,
    D_VALIDATOR_DESC,
    D_VALIDATOR_FEE,
    D_VALIDATOR_ID,
    D_VALIDATOR_MIN_DEL,
    D_VALIDATOR_NAME,
    DEFAULT_DOMAIN_NAME,
    DEFAULT_NODE_NAME,
    DEFAULT_SCHAIN_NAME,
    INITIAL_DELEGATION_PERIOD,
    MONTH_IN_SECONDS,
    SECOND_NODE_NAME,
)
from skale.utils.contracts_provision.utils import (
    generate_random_node_data,
    generate_random_schain_data,
)

DEFAULT_MINING_INTERVAL = 1000
TEST_SRW_FUND_VALUE = 1000000000000000000


def _skip_evm_time(web3: Web3, seconds: int, mine: bool = True) -> int:
    """For test purposes only, works only with hardhat node"""
    res = web3.provider.make_request(RPCEndpoint('evm_increaseTime'), [seconds])
    if mine:
        web3.provider.make_request(RPCEndpoint('evm_mine'), [])
    return int(res['result'])


def set_automining(web3: Web3, value: bool) -> int:
    res = web3.provider.make_request(RPCEndpoint('evm_setAutomine'), [value])
    return int(res['result'])


def set_mining_interval(web3: Web3, ms: int) -> None:
    web3.provider.make_request(RPCEndpoint('evm_setIntervalMining'), [ms])


def set_default_mining_interval(web3: Web3) -> None:
    set_mining_interval(web3, DEFAULT_MINING_INTERVAL)


def add_test_permissions(skale: SkaleManager) -> None:
    add_all_permissions(skale, skale.wallet.address)


def add_all_permissions(skale: SkaleManager, address: ChecksumAddress) -> None:
    default_admin_role = skale.manager.default_admin_role()
    if not skale.manager.has_role(default_admin_role, address):
        skale.manager.grant_role(default_admin_role, address)

    schain_creator_role = skale.schains.schain_creator_role()
    if not skale.schains.has_role(schain_creator_role, address):
        skale.schains.grant_role(schain_creator_role, address)

    schain_removal_role = skale.manager.schain_removal_role()
    if not skale.manager.has_role(schain_removal_role, address):
        skale.manager.grant_role(schain_removal_role, address)

    bounty_reduction_manager_role = skale.bounty_v2.bounty_reduction_manager_role()
    if not skale.bounty_v2.has_role(bounty_reduction_manager_role, address):
        skale.bounty_v2.grant_role(bounty_reduction_manager_role, address)

    locker_manager_role = skale.token_state.locker_manager_role()
    if not skale.token_state.has_role(locker_manager_role, address):
        skale.token_state.grant_role(locker_manager_role, address)

    schain_type_manager_role = skale.schains_internal.schain_type_manager_role()
    if not skale.schains_internal.has_role(schain_type_manager_role, address):
        skale.schains_internal.grant_role(schain_type_manager_role, address)

    validator_manager_role = skale.validator_service.validator_manager_role()
    if not skale.validator_service.has_role(validator_manager_role, address):
        skale.validator_service.grant_role(validator_manager_role, address)

    node_manager_role = skale.nodes.node_manager_role()
    if not skale.nodes.has_role(node_manager_role, address):
        skale.nodes.grant_role(node_manager_role, address)

    compliance_role = skale.nodes.compliance_role()
    if not skale.nodes.has_role(compliance_role, address):
        skale.nodes.grant_role(compliance_role, address)

    constants_holder_role = skale.constants_holder.constants_holder_role()
    if not skale.constants_holder.has_role(constants_holder_role, address):
        skale.constants_holder.grant_role(constants_holder_role, address)

    debugger_role = skale.node_rotation.debugger_role()
    if not skale.node_rotation.has_role(debugger_role, address):
        skale.node_rotation.grant_role(debugger_role, address)

    debugger_schains_role = skale.schains_internal.debugger_role()
    if not skale.schains_internal.has_role(debugger_schains_role, address):
        skale.schains_internal.grant_role(debugger_schains_role, address)

    generation_manager_role = skale.schains_internal.generation_manager_role()
    if not skale.schains_internal.has_role(generation_manager_role, address):
        skale.schains_internal.grant_role(generation_manager_role, address)

    forgiver_role = skale.punisher.forgiver_role()
    if not skale.punisher.has_role(forgiver_role, address):
        skale.punisher.grant_role(forgiver_role, address)

    delegation_period_setter_role = skale.delegation_period_manager.delegation_period_setter_role()
    if not skale.delegation_period_manager.has_role(delegation_period_setter_role, address):
        skale.delegation_period_manager.grant_role(delegation_period_setter_role, address)

    penalty_setter_role = skale.slashing_table.penalty_setter_role()
    if not skale.slashing_table.has_role(penalty_setter_role, address):
        skale.slashing_table.grant_role(penalty_setter_role, address)


def add_test2_schain_type(skale: SkaleManager) -> TxRes:
    part_of_node = 1
    number_of_nodes = 2
    return skale.schains_internal.add_schain_type(part_of_node, number_of_nodes)


def add_test4_schain_type(skale: SkaleManager) -> TxRes:
    part_of_node = 1
    number_of_nodes = 4
    return skale.schains_internal.add_schain_type(part_of_node, number_of_nodes)


def cleanup_nodes(skale: SkaleManager, ids: list[NodeId] | None = None) -> None:
    active_ids = filter(
        lambda i: skale.nodes.node_status(i) == NodeStatus.ACTIVE,
        ids or skale.nodes.active_node_ids(),
    )
    for node_id in active_ids:
        if skale.nodes.get(node_id):
            skale.nodes.init_exit(node_id)
            skale.manager.node_exit(node_id)


def cleanup_schains(skale: SkaleManager) -> None:
    for schain_hash in skale.schains_internal.all_schain_hashes():
        schain_data = skale.schains.get(schain_hash)
        schain_name = schain_data.name
        if schain_name is not None:
            skale.manager.delete_schain_by_root(schain_name, wait_for=True)


def cleanup_nodes_schains(skale: SkaleManager) -> None:
    print('Cleanup nodes and schains')
    cleanup_schains(skale)
    cleanup_nodes(skale)


def create_clean_schain(skale: SkaleManager) -> str:
    cleanup_nodes_schains(skale)
    create_nodes([skale])
    return create_schain(skale, random_name=True)


def create_node(skale: SkaleManager) -> str:
    cleanup_nodes_schains(skale)
    ip, public_ip, port, name = generate_random_node_data()
    skale.manager.create_node(
        ip=ip,
        port=port,
        name=name,
        domain_name=DEFAULT_DOMAIN_NAME,
        public_ip=public_ip,
        wait_for=True,
    )
    return name


def validator_exist(skale: SkaleManager) -> bool:
    return skale.validator_service.validator_address_exists(skale.wallet.address)


def add_delegation_period(skale: SkaleManager) -> None:
    is_added = skale.delegation_period_manager.is_delegation_period_allowed(D_DELEGATION_PERIOD)
    if not is_added:
        skale.delegation_period_manager.set_delegation_period(
            months_count=D_DELEGATION_PERIOD, stake_multiplier=D_STAKE_MULTIPLIER, wait_for=True
        )


def setup_validator(skale: SkaleManager) -> int:
    """Create and activate a validator"""
    set_test_msr(skale)
    print('Address', skale.wallet.address)
    if not validator_exist(skale):
        create_validator(skale)
    else:
        print('Skipping default validator creation')
    validator_id = skale.validator_service.validator_id_by_address(skale.wallet.address)
    if not skale.validator_service.get(validator_id)['trusted']:
        enable_validator(skale, validator_id)
    delegate_to_validator(skale, validator_id)
    delegations = skale.delegation_controller.get_all_delegations_by_validator(validator_id)
    accept_pending_delegation(skale, delegations[-1]['id'])
    _skip_evm_time(skale.web3, MONTH_IN_SECONDS)
    return validator_id


def link_address_to_validator(skale: SkaleManager) -> None:
    print('Linking address to validator')
    signature = skale.validator_service.get_link_node_signature(D_VALIDATOR_ID)
    tx_res = skale.validator_service.link_node_address(
        node_address=skale.wallet.address, signature=signature, wait_for=True
    )
    tx_res.raise_for_status()


def link_nodes_to_validator(
    skale: SkaleManager,
    validator_id: ValidatorId,
    node_skale_objs: Tuple[SkaleManager] | None = None,
) -> None:
    print('Linking address to validator')
    node_skale_objs = node_skale_objs or (skale,)
    validator_id = validator_id or D_VALIDATOR_ID
    for node_skale in node_skale_objs:
        signature = node_skale.validator_service.get_link_node_signature(validator_id)
        skale.validator_service.link_node_address(
            node_address=node_skale.wallet.address, signature=signature
        )


def skip_delegation_delay(skale: SkaleManager, delegation_id: int) -> None:
    print(f'Activating delegation with ID {delegation_id}')
    skale.token_state._skip_transition_delay(delegation_id, wait_for=True)


def accept_pending_delegation(skale: SkaleManager, delegation_id: int) -> None:
    print(f'Accepting delegation with ID: {delegation_id}')
    skale.delegation_controller.accept_pending_delegation(
        delegation_id=delegation_id, wait_for=True
    )


def get_test_delegation_amount(skale: SkaleManager) -> int:
    msr = skale.constants_holder.msr()
    return msr * 30


def set_test_msr(skale: SkaleManager, msr: int = D_VALIDATOR_MIN_DEL) -> None:
    skale.constants_holder._set_msr(new_msr=msr, wait_for=True)


def set_test_mda(skale: SkaleManager) -> None:
    skale.validator_service.set_validator_mda(0, wait_for=True)


def delegate_to_validator(skale: SkaleManager, validator_id: int = D_VALIDATOR_ID) -> None:
    print(f'Delegating tokens to validator ID: {validator_id}')
    skale.delegation_controller.delegate(
        validator_id=validator_id,
        amount=get_test_delegation_amount(skale),
        delegation_period=INITIAL_DELEGATION_PERIOD,
        info=D_DELEGATION_INFO,
        wait_for=True,
    )


def enable_validator(skale: SkaleManager, validator_id: int = D_VALIDATOR_ID) -> None:
    print(f'Enabling validator ID: {D_VALIDATOR_ID}')
    skale.validator_service._enable_validator(validator_id, wait_for=True)


def create_validator(skale: SkaleManager) -> None:
    print('Creating default validator')
    skale.validator_service.register_validator(
        name=D_VALIDATOR_NAME,
        description=D_VALIDATOR_DESC,
        fee_rate=D_VALIDATOR_FEE,
        min_delegation_amount=D_VALIDATOR_MIN_DEL,
        wait_for=True,
    )


def create_nodes(skales: List[SkaleManager], names: List[str] | None = None) -> list[NodeId]:
    # create couple of nodes
    print('Creating two nodes')
    node_names = names or (DEFAULT_NODE_NAME, SECOND_NODE_NAME)
    for skale, name in zip(skales, node_names):
        ip, public_ip, port, _ = generate_random_node_data()
        skale.manager.create_node(
            ip=ip,
            port=port,
            name=name,
            domain_name=DEFAULT_DOMAIN_NAME,
            public_ip=public_ip,
            wait_for=True,
        )
    ids = [skales[0].nodes.node_name_to_index(name) for name in node_names]
    return ids


def create_schain(
    skale: SkaleManager,
    schain_name: SchainName = DEFAULT_SCHAIN_NAME,
    schain_type: int = 1,
    random_name: bool = False,
    schain_options: SchainOptions | None = None,
) -> str:
    print('Creating schain')
    # create 1 s-chain
    type_of_nodes, lifetime_seconds, name = generate_random_schain_data(skale)

    if random_name:
        schain_name = name

    if not schain_type:
        schain_type = type_of_nodes

    skale.schains.add_schain_by_foundation(
        lifetime_seconds,
        schain_type,
        0,
        schain_name,
        options=schain_options,
        wait_for=True,
        value=TEST_SRW_FUND_VALUE,
    )
    return schain_name



"""SKALE contract manager test"""

from skale_contracts.projects.skale_manager import SkaleManagerContract
from web3 import Web3

from skale.utils.constants import ZERO_ADDRESS
from tests.constants import TEST_CONTRACT_NAME, TEST_CONTRACT_NAME_HASH


def test_get_contract_address(skale):
    contract_address = skale.instance.get_contract_address(SkaleManagerContract.NODES)
    assert Web3.to_checksum_address(contract_address) != ZERO_ADDRESS


def test_get_contract_hash_by_name(skale):
    contract_name_hash = skale.contract_manager.get_contract_hash_by_name(TEST_CONTRACT_NAME)
    assert contract_name_hash == TEST_CONTRACT_NAME_HASH



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import enum
from typing import Mapping, NamedTuple

from eth_typing import HexStr
from web3.types import TxReceipt

from skale.transactions.exceptions import (
    DryRunFailedError,
    DryRunRevertError,
    TransactionFailedError,
)


class TxStatus(int, enum.Enum):
    FAILED = 0
    SUCCESS = 1


class TxCallResult(NamedTuple):
    status: TxStatus
    error: str
    message: str
    data: Mapping[str, str | int]


class TxRes:
    def __init__(
        self,
        tx_call_result: TxCallResult | None = None,
        tx_hash: HexStr | None = None,
        receipt: TxReceipt | None = None,
    ):
        self.tx_call_result = tx_call_result
        self.tx_hash = tx_hash
        self.receipt = receipt
        self.attempts = 0

    def __str__(self) -> str:
        return (
            f'TxRes hash: {self.tx_hash}, tx_call_result {self.tx_call_result}, '
            f'receipt {self.receipt}'
        )

    def __repr__(self) -> str:
        return (
            f'TxRes hash: {self.tx_hash}, tx_call_result {self.tx_call_result}, '
            f'receipt {self.receipt}'
        )

    def raise_for_status(self) -> None:
        if self.receipt is not None:
            if self.receipt['status'] == TxStatus.FAILED:
                raise TransactionFailedError(
                    'Tx status is failed', {key: str(value) for key, value in self.receipt.items()}
                )
        elif self.tx_call_result is not None and self.tx_call_result.status == TxStatus.FAILED:
            if self.tx_call_result.error == 'revert':
                raise DryRunRevertError(self.tx_call_result.message)
            else:
                raise DryRunFailedError(self.tx_call_result.message)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.schain_config import PORTS_PER_SCHAIN
from skale.types.node import Port
from skale.types.schain import SchainHash
from skale.utils.exceptions import SChainNotFoundException


def calc_schain_base_port(node_base_port: Port, schain_index: int) -> Port:
    return Port(node_base_port + schain_index * PORTS_PER_SCHAIN)


def get_schain_index_in_node(
    schain_hash: SchainHash, schain_hashes_on_node: list[SchainHash]
) -> int:
    try:
        return schain_hashes_on_node.index(schain_hash)
    except ValueError:
        raise SChainNotFoundException(
            f'sChain {str(schain_hash)} is not found in the list: {str(schain_hashes_on_node)}'
        )


def get_schain_base_port_on_node(
    schain_hashes_on_node: list[SchainHash], schain_hash: SchainHash, node_base_port: Port
) -> Port:
    schain_index = get_schain_index_in_node(schain_hash, schain_hashes_on_node)
    return calc_schain_base_port(node_base_port, schain_index)



import importlib
import os
from datetime import datetime
from unittest import mock

import pytest
from freezegun import freeze_time
from web3.exceptions import StaleBlockchain

import skale.config as config
from skale.utils.exceptions import NoSyncedEndpointError
from skale.utils.web3_utils import get_endpoint
from tests.constants import ENDPOINT


@pytest.fixture
def last_block_file():
    filepath = 'last-block-file'
    os.environ['LAST_BLOCK_FILE'] = filepath
    importlib.reload(config)
    try:
        yield filepath
    finally:
        if os.path.isfile(filepath):
            os.remove(filepath)
            os.environ.pop('LAST_BLOCK_FILE')
        importlib.reload(config)


def test_call_with_outdated_client(skale):
    # because of skipTime in preparation
    current_ts = skale.web3.eth.get_block('latest')['timestamp']
    allowed_diff = config.ALLOWED_TS_DIFF
    dt = datetime.utcfromtimestamp(current_ts + allowed_diff)
    with freeze_time(dt):
        skale.validator_service.ls()
    dt = datetime.utcfromtimestamp(current_ts + allowed_diff + 15)
    with freeze_time(dt):
        with pytest.raises(StaleBlockchain):
            skale.validator_service.ls()


def test_transaction_with_outdated_client(skale):
    # because of skipTime in preparation
    current_ts = skale.web3.eth.get_block('latest')['timestamp']
    allowed_diff = config.ALLOWED_TS_DIFF

    dt = datetime.utcfromtimestamp(current_ts + allowed_diff)
    new_rotation_delay = 100
    with freeze_time(dt):
        skale.constants_holder.set_rotation_delay(new_rotation_delay, wait_for=True)

    dt = datetime.utcfromtimestamp(current_ts + allowed_diff + 15)
    new_rotation_delay = 101
    with freeze_time(dt):
        with pytest.raises(StaleBlockchain):
            skale.constants_holder.set_rotation_delay(new_rotation_delay, wait_for=True)


def test_get_endpoint():
    with mock.patch('web3.main.Web3.is_connected', return_value=True):
        endpoint = get_endpoint([ENDPOINT, 'http://localhost:8546'])
        assert endpoint == ENDPOINT

        endpoint = get_endpoint('http://localhost:1111')
        assert endpoint == 'http://localhost:1111'

    with pytest.raises(NoSyncedEndpointError):
        get_endpoint(['http://incorrect.endpoint'])

    endpoint = get_endpoint(['http://localhost:1111', ENDPOINT])
    assert endpoint == ENDPOINT


def test_get_endpoint_stale_blockchain(web3):
    current_ts = web3.eth.get_block('latest')['timestamp']
    allowed_diff = config.ALLOWED_TS_DIFF
    dt = datetime.utcfromtimestamp(current_ts + allowed_diff)
    with freeze_time(dt):
        endpoint = get_endpoint(['http://localhost:1111', ENDPOINT])
        assert endpoint == ENDPOINT
    dt = datetime.utcfromtimestamp(current_ts + allowed_diff + 15)
    with freeze_time(dt):
        with pytest.raises(NoSyncedEndpointError):
            get_endpoint(['http://localhost:1111', ENDPOINT])
